from ..._common import filetype_from_filename, register_format

__all__ = [
    "register",
    "read",
    "write",
]


_extension_to_filetype = {}
_reader_map = {}
_writer_map = {}
_file_formats = {
    "INFILE": "tough",
    "MESH": "tough",
    "INCON": "tough",
    "GENER": "tough",
    "flow.inp": "toughreact-flow",
    "solute.inp": "toughreact-solute",
    "chemical.inp": "toughreact-chemical",
}


def register(file_format, extensions, reader, writer=None):
    """
    Register a new input format.

    Parameters
    ----------
    file_format : str
        File format to register.
    extensions : array_like
        List of extensions to associate to the new format.
    reader : callable
        Read fumction.
    writer : callable or None, optional, default None
        Write function.

    """
    register_format(
        fmt=file_format,
        ext_to_fmt=_extension_to_filetype,
        reader_map=_reader_map,
        writer_map=_writer_map,
        extensions=extensions,
        reader=reader,
        writer=writer,
    )


def read(filename, file_format=None, **kwargs):
    """
    Read TOUGH input file.

    Parameters
    ----------
    filename : str, pathlike or buffer
        Input file name or buffer.
    file_format : str ('tough', 'toughreact-flow', 'toughreact-solute', 'toughreact-chemical', 'json') or None, optional, default None
        Input file format.

    Other Parameters
    ----------------
    blocks : list of str or None, optional, default None
        Only if ``file_format = "tough"``. Blocks to read. If None, all blocks are read.
    label_length : int or None, optional, default None
        Only if ``file_format = "tough"``. Number of characters in cell labels.
    n_variables : int or None, optional, default None
        Only if ``file_format = "tough"``. Number of primary variables.
    eos : str or None, optional, default None
        Only if ``file_format = "tough"``. Equation of State.
    mopr_11 : int, optional, default 0
        Only if ``file_format = "toughreact-solute"``. MOPR(11) value in file 'flow.inp'.

    Returns
    -------
    dict
        TOUGH input parameters.

    Note
    ----
    If ``file_format == 'tough'``, can also read `MESH`, `INCON` and `GENER` files.

    """
    if not (file_format is None or file_format in _reader_map):
        raise ValueError()

    file_format = _get_file_format(filename, file_format, default="tough")
    return _reader_map[file_format](filename, **kwargs)


def write(filename, parameters, file_format=None, **kwargs):
    """
    Write TOUGH input file.

    Parameters
    ----------
    filename : str, pathlike or buffer
        Output file name or buffer.
    parameters : dict
        Parameters to export.
    file_format : str ('tough', 'toughreact-flow', 'toughreact-solute', 'toughreact-chemical', 'json') or None, optional, default None
        Output file format.

    Other Parameters
    ----------------
    block : str {'all', 'gener', 'mesh', 'incon'} or None, optional, default None
        Only if ``file_format = "tough"``. Blocks to be written:

         - 'all': write all blocks,
         - 'gener': only write block GENER,
         - 'mesh': only write blocks ELEME, COORD and CONNE,
         - 'incon': only write block INCON,
         - None: write all blocks except blocks defined in `ignore_blocks`.

    ignore_blocks : list of str or None, optional, default None
        Only if ``file_format = "tough"`` and `block` is None. Blocks to ignore.
    space_between_blocks : bool, optional, default False
        Only if ``file_format = "tough"``. Add an empty record between blocks.
    space_between_blocks : bool, optional, default True
        Only if ``file_format = "tough"``. Add a white space between floating point values.
    eos : str or None, optional, default None
        Only if ``file_format = "tough"``. Equation of State.
        If `eos` is defined in `parameters`, this option will be ignored.
    mopr_10 : int, optional, default 0
        Only if ``file_format = "toughreact-solute"``. MOPR(10) value in file 'flow.inp'.
    mopr_11 : int, optional, default 0
        Only if ``file_format = "toughreact-solute"``. MOPR(11) value in file 'flow.inp'.
    verbose : bool, optional, default True
        Only if ``file_format`` in {"toughreact-solute", "toughreact-chemical"}. If `True`, add comments to describe content of file.

    """
    if not isinstance(parameters, dict):
        raise TypeError()
    if not (file_format is None or file_format in _writer_map):
        raise ValueError()

    file_format = _get_file_format(filename, file_format, default="tough")
    _writer_map[file_format](filename, parameters, **kwargs)


def _get_file_format(filename, file_format, default):
    """Get file format."""
    if not file_format:
        file_format = _file_format_from_filename(filename)

    if not file_format:
        file_format = filetype_from_filename(filename, _extension_to_filetype, default)

    if not file_format:
        file_format = "tough"

    return file_format


def _file_format_from_filename(filename):
    """Determine file format from its name."""
    import pathlib
    from io import TextIOWrapper

    if not isinstance(filename, TextIOWrapper):
        filename = pathlib.Path(filename).name

        return _file_formats[filename] if filename in _file_formats else ""

    else:
        return ""



import json

import numpy as np

from ...._common import open_file

__all__ = [
    "read",
    "write",
]


def read(filename):
    """
    Import json TOUGH input file.

    Parameters
    ----------
    filename : str, pathlike or buffer
        Input file name or buffer.

    Returns
    -------
    dict
        TOUGH input parameters.

    """

    def to_int(data):
        """Return dict with integer keys instead of strings."""
        return {int(k): data[k] for k in sorted(data)}

    with open_file(filename, "r") as f:
        parameters = json.load(f)

    if "react" in parameters and "options" in parameters["react"]:
        parameters["react"]["options"] = to_int(parameters["react"]["options"])

    for key in {"extra_options", "more_options", "hysteresis_options"}:
        if key in parameters:
            parameters[key] = to_int(parameters[key])

    if "selections" in parameters and "integers" in parameters["selections"]:
        parameters["selections"]["integers"] = to_int(
            parameters["selections"]["integers"]
        )

    return parameters


def write(filename, parameters):
    """
    Export TOUGH parameters to json.

    Parameters
    ----------
    filename : str, pathlike or buffer
        Output file name or buffer.
    parameters : dict
        Parameters to export.

    """
    from copy import deepcopy

    def jsonify(x):
        """JSON serialize data."""
        if isinstance(x, (np.int32, np.int64)):
            return int(x)
        elif isinstance(x, (list, tuple)):
            return [jsonify(xx) for xx in x]
        elif isinstance(x, np.ndarray):
            return x.tolist()
        elif isinstance(x, dict):
            return {k: jsonify(v) for k, v in x.items()}
        else:
            return x

    with open_file(filename, "w") as f:
        parameters = deepcopy(parameters)
        parameters = jsonify(parameters)
        json.dump(parameters, f, indent=4)



"""
Generate input files
====================

The TOUGH input file is shown below and specifies a cylindrical heater of 0.3 m radius and 4.5 m height, that provides a constant output of 3 kW into a porous medium with uniform initial conditions of 18°C temperature, 1 bar pressure, and 20% gas saturation. The mesh consists of a one-dimensional radial grid of 120 active elements extending to a radius of 10,000 m (practically infinite for the time scales of interest here), with an additional inactive element of zero volume representing constant boundary conditions. Properly speaking, the problem represents one unit of an infinite linear string of identical heaters; if a single heater were to be modeled, important end effects would occur at the top and bottom, and a two-dimensional RZ grid would have to be used.

Most of the formation parameters are identical to data used in previous modeling studies of high-level nuclear waste emplacement at Yucca Mountain (Pruess et al., 1990). As we do not include fracture effects in the present simulation, heat pipe effects would be very weak at the low rock matrix permeabilities (of order 1 microdarcy) encountered at Yucca Mountain. To get a more interesting behavior, we have arbitrarily increased absolute permeability by something like a factor 10,000, to 20 millidarcy, and for consistency have reduced capillary pressures by a factor (10,000)\ :sup:`1/2` = 100 in comparison to typical Yucca Mountain data.

.. code-block::

    *rhp* 1-D RADIAL HEAT PIPE
    MESHMAKER1----*----2----*----3----*----4----*----5----*----6----*----7----*----8
    RZ2D
    RADII
        1
            0.
    EQUID
        1             .3
    LOGAR
    99           1.E2
    LOGAR
    20           1.E4
    EQUID
        1            0.0
    LAYER----1----*----2----*----3----*----4----*----5----*----6----*----7----*----8
        1
        4.5
    
    ROCKS----1----*----2----*----3----*----4----*----5----*----6----*----7----*----8
    POMED    1     2550.       .10   20.E-15   20.E-15   20.E-15       2.0     800.0
                                         .25

    MULTI----1----*----2----*----3----*----4----*----5----*----6----*----7----*----8
        2    3    2    6
    START----1----*----2----*----3----*----4----*----5----*----6----*----7----*----8
    ----*----1 MOP: 123456789*123456789*1234 ---*----5----*----6----*----7----*----8
    PARAM----1----*----2----*----3----*----4----*----5----*----6----*----7----*----8
    2 250     25000003000000002 47 1 1                   1.80
               3.15576E8       -1.
          1.E3      9.E3      9.E4      4.E5
         1.E-5     1.E00                                   1.E-7
                    1.E5                0.20                 18.
    RPCAP----1----*----2----*----3----*----4----*----5----*----6----*----7----*----8
        7        0.45000    9.6E-4        1.
        7        0.45000    1.0E-3   8.0E-05      5.E8        1.
    TIMES----1----*----2----*----3----*----4----*----5----*----6----*----7----*----8
        3
     3.15576E7  1.2559E8 3.15576E8
    GENER----1----*----2----*----3----*----4----*----5----*----6----*----7----*----8
    A1  1HTR 1                         HEAT       3.E3
    
    ENDCY----1----*----2----*----3----*----4----*----5----*----6----*----7----*----8


This example shows how to generate this file from scratch block by block using :mod:`toughio`.

"""

########################################################################################
# First, we import :mod:`numpy` and :mod:`toughio`.

import numpy as np
import toughio

########################################################################################

########################################################################################
# In this example, we start with the blocks MESHM and LAYER, and use the function :func:`toughio.meshmaker.from_meshmaker` to generate the radial grid. We first define a dictionary with the key "meshmaker" and provide the type of mesh (RZ2D) and the parameters. Once the mesh is generated, we use the method :meth:`toughio.Mesh.write_tough` and :meth:`toughio.Mesh.write` to write the files `MESH` and `mesh.pickle`, respectively. Exporting the mesh is useful for postprocessing as it contains information about the mesh.

parameters = {
    "meshmaker": {
        "type": "rz2d",
        "parameters": [
            {"type": "radii", "radii": [0.0]},
            {"type": "equid", "n_increment": 1, "size": 0.3},
            {"type": "logar", "n_increment": 99, "radius": 100.0},
            {"type": "logar", "n_increment": 20, "radius": 10000.0},
            {"type": "equid", "n_increment": 1, "size": 0.0},
            {"type": "layer", "thicknesses": [4.5]},
        ]
    }
}
mesh = toughio.meshmaker.from_meshmaker(parameters, material="POMED")
mesh.write_tough("MESH")
mesh.write("mesh.pickle")

########################################################################################

########################################################################################
# To generate the main input file, let us initialize a new dictionary with a title (optional).

parameters = {
    "title": "*rhp* 1-D RADIAL HEAT PIPE",
}

########################################################################################

########################################################################################
# The block ROCKS contains a single material and can be defined using the key "rocks", which is a dictionary with keys corresponding to the names of the materials.

parameters["rocks"] = {
    "POMED": {
        "density": 2550.0,
        "porosity": 0.1,
        "permeability": 20.0e-15,
        "conductivity": 2.0,
        "specific_heat": 800.0,
        "tortuosity": 0.25,
    },
}

########################################################################################

########################################################################################
# We can initialize the block MULTI by setting the equation-of-state (EOS3) and the isothermal option. The key "eos" defines the number of mass components and phases while the isothermal option determines the number of equations (i.e., the number of equations is equal to the number of mass components for isothermal simulations). Alternatively, the number of mass components and phases can be fixed using the keys "n_component" and "n_phase", respectively.

parameters["eos"] = "eos3"
parameters["isothermal"] = False

########################################################################################

########################################################################################
# The block START allows INCON data to be defined in arbitrary order and not present for all grid blocks. This option can be enable using the key "start".

parameters["start"] = True

########################################################################################

########################################################################################
# The block PARAM is mainly defined by the key "options". However, the MOP values and the default initial conditions for all grid blocks (last record of the block PARAM) are set in the keys "extra_options" and "default", respectively.

parameters["options"] = {
    "verbosity": 2,
    "n_cycle": 250,
    "n_cycle_print": 250,
    "temperature_dependence_gas": 1.8,
    "t_max": 10.0 * 24.0 * 3600.0 * 365.25,
    "t_steps": [1.0e3, 9.0e3, 9.0e4, 4.0e5],
    "eps1": 1.0e-5,
    "eps2": 1.0,
    "derivative_factor": 1.0e-7,
}
parameters["extra_options"] = {
    1: 1,
    5: 3,
    16: 4,
    17: 7,
    19: 1,
    21: 6,
}
parameters["default"] = {
    "initial_condition": [1.0e5, 0.2, 18.0],
}

########################################################################################

########################################################################################
# The block RPCAP controls the default relative permeability and capillarity models for all materials. Therefore, they are also defined in the dictionary key "default".

parameters["default"]["relative_permeability"] = {
    "id": 7,
    "parameters": [0.45, 0.00096, 1.0],
}
parameters["default"]["capillarity"] = {
    "id": 7,
    "parameters": [0.45, 0.001, 8.0e-5, 5.0e8, 1.0],
}

########################################################################################

########################################################################################
# The block TIMES is simply a list of values (in second) defined using the homologous key "times". Here, we request an output after 1, 4 and 10 years.

parameters["times"] = np.array([1.0, 4.0, 10.0]) * 24.0 * 3600.0 * 365.25

########################################################################################

########################################################################################
# The block GENER is defined by the key "generators" which is a list of dictionaries with keys "label", "name" (optional), "type", "times", "rates" and "specific_enthalpy". "label" is the name of the element where the component is injected/produced. Here, we use the method :meth:`toughio.Mesh.near` to get the index of the element the nearest to the query point.

parameters["generators"] = [
    {
        "label": mesh.labels[mesh.near([0.0, 0.0, 0.0])],
        "name": "HTR 1",
        "type": "HEAT",
        "rates": 3.0e3,
    },
]

########################################################################################

########################################################################################
# Once all the parameters are defined, the input file is ready to written using the function :func:`toughio.write_input`.

toughio.write_input("INFILE", parameters)

########################################################################################



import os
import platform
import sys

import helpers
import pytest

import toughio


@pytest.mark.parametrize(
    "exec, workers, docker, wsl, cmd",
    [
        ("tough-exec", None, None, False, "tough-exec INFILE INFILE.out"),
        ("tough-exec", 8, None, False, "mpiexec -n 8 tough-exec INFILE INFILE.out"),
    ],
)
def test_run(exec, workers, docker, wsl, cmd):
    toughio._run._check_exec = False
    status = toughio.run(
        exec,
        {},
        workers=workers,
        docker=docker,
        wsl=wsl,
        working_dir=helpers.tempdir(),
        # dirs_exist_ok only works with Python > 3.8
        use_temp=sys.version_info >= (3, 8),
        ignore_patterns=["INFILE"],
        silent=True,
    )

    assert status.args == cmd


@pytest.mark.skipif(
    not platform.system().startswith("Win"), reason="requires Windows platform"
)
@pytest.mark.parametrize(
    "exec, workers, docker, wsl, cmd",
    [
        (
            "tough-exec",
            None,
            "docker-image",
            False,
            "docker run --rm  -v PLACEHOLDER:/shared -w /shared docker-image tough-exec INFILE INFILE.out",
        ),
        ("tough-exec", None, None, True, "bash -c 'tough-exec INFILE INFILE.out'"),
        (
            "tough-exec",
            8,
            "docker-image",
            True,
            """bash -c 'docker run --rm  -v PLACEHOLDER:/shared -w /shared docker-image mpiexec -n 8 tough-exec INFILE INFILE.out'""",
        ),
    ],
)
def test_run_windows(exec, workers, docker, wsl, cmd):
    if os.getenv("ComSpec").endswith("cmd.exe"):
        cmd = cmd.replace("PLACEHOLDER", '"%cd%"')

    else:
        cmd = cmd.replace("PLACEHOLDER", "${PWD}")

    test_run(exec, workers, docker, wsl, cmd)


@pytest.mark.skipif(
    platform.system().startswith("Win"), reason="requires Unix platform"
)
@pytest.mark.parametrize(
    "exec, workers, docker, cmd",
    [
        (
            "tough-exec",
            None,
            "docker-image",
            "docker run --rm PLACEHOLDER -v ${PWD}:/shared -w /shared docker-image tough-exec INFILE INFILE.out",
        ),
        (
            "tough-exec",
            8,
            "docker-image",
            "docker run --rm PLACEHOLDER -v ${PWD}:/shared -w /shared docker-image mpiexec -n 8 tough-exec INFILE INFILE.out",
        ),
    ],
)
def test_run_unix(exec, workers, docker, cmd):
    cmd = cmd.replace("PLACEHOLDER", f"-e LOCAL_USER_ID={os.getuid()}")
    test_run(exec, workers, docker, False, cmd)



import numpy as np

from ...._common import open_file
from .._common import to_output

__all__ = [
    "read",
    "write",
]


zone_key_to_type = {
    "T": str,
    "I": int,
    "J": int,
    "K": int,
    "C": str,
    "N": int,
    "NODES": int,
    "E": int,
    "ELEMENTS": int,
    "F": str,
    "ET": str,
    "DATAPACKING": str,
    "ZONETYPE": str,
    "NV": int,
    "VARLOCATION": str,
}


def read(filename, file_type, labels_order=None):
    """
    Read OUTPUT_ELEME.tec.

    Parameters
    ----------
    filename : str, pathlike or buffer
        Input file name or buffer.
    file_type : str
        Input file type.
    labels_order : list of array_like
        List of labels. If None, output will be assumed ordered.

    Returns
    -------
    namedtuple or list of namedtuple
        namedtuple (type, format, time, labels, data) or list of namedtuple for each time step.

    """
    with open_file(filename, "r") as f:
        headers, zones = read_buffer(f)

    times, labels, variables = [], [], []
    for zone in zones:
        time = float(zone["title"].split()[0]) if "title" in zone else None

        times.append(time)
        labels.append([])
        variables.append(zone["data"])

    return to_output(file_type, labels_order, headers, times, labels, variables)


def read_buffer(f):
    """Read OUTPUT_ELEME.tec."""
    zones = []

    # Loop until end of file
    while True:
        line = f.readline().strip()

        # Read header (VARIABLES)
        if line.upper().startswith("VARIABLES"):
            headers = _read_variables(line)

        # Read zone
        elif line.upper().startswith("ZONE"):
            zone = _read_zone(line)

            if "I" not in zone:
                raise ValueError()

            # Read data
            data = np.genfromtxt(f, max_rows=zone["I"])

            # Output
            tmp = {"data": data}
            tmp["title"] = zone["T"]
            zones.append(tmp)

        elif not line:
            break

    return headers, zones


def write(filename, output):
    """
    Write OUTPUT_ELEME.tec.

    Parameters
    ----------
    filename : str, pathlike or buffer
        Output file name or buffer.
    output : namedtuple or list of namedtuple
        namedtuple (type, format, time, labels, data) or list of namedtuple for each time step to export.

    """
    out = output[-1]
    headers = []

    if "X" in out.data:
        headers += ["X"]

    else:
        raise ValueError()

    if "Y" in out.data:
        headers += ["Y"]

    else:
        raise ValueError()

    headers += ["Z"] if "Z" in out.data else []
    headers += [k for k in out.data if k not in {"X", "Y", "Z"}]

    with open_file(filename, "w") as f:
        # Headers
        record = "".join(f"{header:>18}" for header in headers)
        f.write(f" VARIABLES       ={record}\n")

        # Data
        for out in output:
            # Zone
            record = f' ZONE T="{out.time:14.7e} SEC"  I = {len(out.data["X"]):8d}'
            f.write(f"{record}\n")

            # Table
            data = np.transpose([out.data[k] for k in headers])
            for d in data:
                record = "".join(f"{x:20.12e}" for x in d)
                f.write(f"{' ' * 18}{record}\n")


def _read_variables(line):
    # Gather variables in a list
    line = line.split("=")[1]
    line = [x for x in line.replace(",", " ").split()]
    variables = []

    i = 0
    while i < len(line):
        if '"' in line[i] and not (line[i].startswith('"') and line[i].endswith('"')):
            var = f"{line[i]} {line[i + 1]}"
            i += 1

        else:
            var = line[i]

        variables.append(var.replace('"', ""))
        i += 1

    return [variable for variable in variables if variable]


def _read_zone(line):
    # Gather zone entries in a dict
    line = line[5:]
    zone = {}

    # Look for zone title
    ivar = line.find('"')

    # If zone contains a title, process it and save the title
    if ivar >= 0:
        i1, i2 = ivar, ivar + line[ivar + 1 :].find('"') + 2
        zone_title = line[i1 + 1 : i2 - 1]
        line = line.replace(line[i1:i2], "PLACEHOLDER")

    else:
        zone_title = None

    # Look for VARLOCATION (problematic since it contains both ',' and '=')
    ivar = line.find("VARLOCATION")

    # If zone contains VARLOCATION, process it and remove the key/value pair
    if ivar >= 0:
        i1, i2 = line.find("("), line.find(")")
        zone["VARLOCATION"] = line[i1 : i2 + 1].replace(" ", "")
        line = line[:ivar] + line[i2 + 1 :]

    # Split remaining key/value pairs separated by '='
    line = [x for x in line.replace(",", " ").split() if x != "="]
    i = 0
    while i < len(line) - 1:
        if "=" in line[i]:
            if not (line[i].startswith("=") or line[i].endswith("=")):
                key, value = line[i].split("=")

            else:
                key = line[i].replace("=", "")
                value = line[i + 1]
                i += 1

        else:
            key = line[i]
            value = line[i + 1].replace("=", "")
            i += 1

        zone[key] = zone_key_to_type[key](value)
        i += 1

    # Add zone title to zone dict
    if zone_title:
        zone["T"] = zone_title.strip()

    return zone



__all__ = [
    "merge",
]


def merge(argv=None):
    import os

    parser = _get_parser()
    args = parser.parse_args(argv)

    # Check that input, MESH and INCON files exist
    head = os.path.split(args.infile)[0]
    gener_filename = head + ("/" if head else "") + "GENER"
    mesh_filename = head + ("/" if head else "") + "MESH"
    incon_filename = head + ("/" if head else "") + "INCON"

    if not os.path.isfile(args.infile):
        raise ValueError(f"File '{args.infile}' not found.")

    gener_exists = os.path.isfile(gener_filename)
    mesh_exists = os.path.isfile(mesh_filename)
    incon_exists = os.path.isfile(incon_filename)

    if not (gener_exists or mesh_exists or incon_exists):
        raise ValueError("GENER, MESH or INCON file(s) not found. Nothing to merge.")

    # Buffer input file
    with open(args.infile, "r") as f:
        input_file = list(f)

    # Check that input file has at least blocks ROCKS, PARAM, ENDFI or ENDCY
    count = 0
    for line in input_file:
        count += int(line.upper()[:5] in {"ROCKS", "PARAM", "ENDFI", "ENDCY"})
    if count < 3:
        raise ValueError(f"Invalid input file '{args.infile}'.")

    # Buffer GENER
    if gener_exists:
        gener_file = _read_file(gener_filename, end="+++")

        if not gener_file[0].startswith("GENER"):
            raise ValueError("Invalid GENER file.")

    # Buffer MESH
    if mesh_exists:
        mesh_file = _read_file(mesh_filename)

        if not mesh_file[0].startswith("ELEME"):
            raise ValueError("Invalid MESH file.")

    # Buffer INCON if exist
    if incon_exists:
        incon_file = _read_file(incon_filename, end="+++")

        if not incon_file[0].startswith("INCON"):
            raise ValueError("Invalid INCON file.")

    # Locate ENDFI or ENDCY
    for i, line in enumerate(input_file):
        if line.upper()[:5] in {"ENDFI", "ENDCY"}:
            break

    # Buffer output file
    output_file = input_file[:i]

    if gener_exists:
        output_file += gener_file

    if mesh_exists:
        output_file += mesh_file

    if incon_exists:
        output_file += incon_file

    output_file += input_file[i:]

    # Write output file
    with open(args.outfile, "w") as f:
        for line in output_file:
            f.write(line)


def _get_parser():
    import argparse

    # Initialize parser
    parser = argparse.ArgumentParser(
        description=(
            "Merge input file, GENER and/or MESH and/or INCON into a single file. "
            "The files must be in the same directory as input file."
        ),
        formatter_class=argparse.RawTextHelpFormatter,
    )

    # Input file
    parser.add_argument(
        "infile",
        type=str,
        help="TOUGH input file",
    )

    # Output file
    parser.add_argument("outfile", type=str, help="merged TOUGH input file")

    return parser


def _read_file(filename, end=None):
    with open(filename, "r") as f:
        if end is None:
            file = list(f)

        else:
            file = []

            for line in f:
                if line.startswith(end):
                    file.append("\n")
                    break

                file.append(line)

    return file



from ...._common import block_to_format, get_label_length, open_file, prune_values
from ...._exceptions import ReadError
from ...._helpers import FileIterator
from ..._common import read_record
from .._common import read_end_comments
from ._helpers import read_model_record, read_primary_variables

__all__ = [
    "read",
]


oft_to_key = {
    "FOFT": "element_history",
    "COFT": "connection_history",
    "GOFT": "generator_history",
}


def read(
    filename,
    blocks=None,
    label_length=None,
    n_variables=None,
    eos=None,
    simulator="tough",
):
    """
    Read TOUGH input file.

    Parameters
    ----------
    filename : str, pathlike or buffer
        Input file name or buffer.
    blocks : list of str or None, optional, default None
        Blocks to read. If None, all blocks are read.
    label_length : int or None, optional, default None
        Number of characters in cell labels.
    n_variables : int or None, optional, default None
        Number of primary variables.
    eos : str or None, optional, default None
        Equation of State.

    Returns
    -------
    dict
        TOUGH input parameters.

    """
    if not (label_length is None or isinstance(label_length, int)):
        raise TypeError()
    if isinstance(label_length, int) and not 5 <= label_length < 10:
        raise ValueError()
    if simulator not in {"tough", "toughreact"}:
        raise ValueError()

    with open_file(filename, "r") as f:
        out = read_buffer(f, blocks, label_length, n_variables, eos, simulator)

    return out


def read_buffer(f, block_stack, label_length, n_variables, eos, simulator="tough"):
    """Read TOUGH input file."""
    from ._common import blocks

    # Block filters
    block_stack = block_stack if block_stack is not None else blocks
    block_stack = set(block_stack)

    parameters = {}
    flag = False

    # Title
    if "TITLE" in block_stack:
        block_stack.remove("TITLE")

        title = []
        while True:
            if len(title) >= 100:
                raise ValueError()

            line = f.readline().strip()

            if line[:5].rstrip().upper() not in blocks:
                title.append(line)

            else:
                break

        if title:
            title = title[0] if len(title) == 1 else title

        if title:
            parameters["title"] = title

        f.seek(0)

    # Loop over blocks
    # Some blocks (INCON, INDOM, PARAM) need to rewind to previous line but tell and seek are disabled by next
    # See <https://stackoverflow.com/questions/22688505/is-there-a-way-to-go-back-when-reading-a-file-using-seek-and-calls-to-next>
    fiter = FileIterator(f)

    try:
        for line in fiter:
            if line.startswith("DIMEN") and "DIMEN" in block_stack:
                block_stack.remove("DIMEN")
                parameters.update(_read_dimen(fiter))

            elif line.startswith("ROCKS") and "ROCKS" in block_stack:
                block_stack.remove("ROCKS")
                parameters.update(_read_rocks(fiter, simulator))

            elif line.startswith("RPCAP") and "RPCAP" in block_stack:
                block_stack.remove("RPCAP")
                rpcap = _read_rpcap(fiter)

                if "default" in parameters:
                    parameters["default"].update(rpcap)

                else:
                    parameters["default"] = rpcap

            elif line.startswith("REACT") and "REACT" in block_stack:
                block_stack.remove("REACT")
                react = _read_react(fiter)

                if "react" in parameters:
                    parameters["react"].update(react["react"])

                else:
                    parameters.update(react)

            elif line.startswith("FLAC") and "FLAC" in block_stack:
                block_stack.remove("FLAC")
                flac = _read_flac(fiter, parameters["rocks"])
                parameters["flac"] = flac["flac"]

                for k, v in flac["rocks"].items():
                    parameters["rocks"][k].update(v)

            elif line.startswith("CHEMP") and "CHEMP" in block_stack:
                block_stack.remove("CHEMP")
                parameters.update(_read_chemp(fiter))

            elif line.startswith("NCGAS") and "NCGAS" in block_stack:
                block_stack.remove("NCGAS")
                parameters.update(_read_ncgas(fiter))

            elif line.startswith("MULTI") and "MULTI" in block_stack:
                block_stack.remove("MULTI")
                parameters.update(_read_multi(fiter))

                if not n_variables:
                    n_variables = parameters["n_component"] + 1

            elif line.startswith("SOLVR") and "SOLVR" in block_stack:
                block_stack.remove("SOLVR")
                parameters.update(_read_solvr(fiter))

            elif line.startswith("INDEX") and "INDEX" in block_stack:
                block_stack.remove("INDEX")
                parameters["index"] = True

            elif line.startswith("START") and "START" in block_stack:
                block_stack.remove("START")
                parameters["start"] = True

            elif line.startswith("PARAM") and "PARAM" in block_stack:
                block_stack.remove("PARAM")
                param, n_variables = _read_param(fiter, n_variables, eos)
                parameters["options"] = param["options"]
                parameters["extra_options"] = param["extra_options"]

                if "default" in parameters:
                    parameters["default"].update(param["default"])

                else:
                    parameters["default"] = param["default"]

            elif line.startswith("SELEC") and "SELEC" in block_stack:
                block_stack.remove("SELEC")
                parameters.update(_read_selec(fiter))

            elif line.startswith("INDOM") and "INDOM" in block_stack:
                block_stack.remove("INDOM")
                indom, n_variables = _read_indom(fiter, n_variables, eos)

                for k, v in indom["rocks"].items():
                    parameters["rocks"][k].update(v)

            elif line.startswith("MOMOP") and "MOMOP" in block_stack:
                block_stack.remove("MOMOP")
                parameters.update(_read_momop(fiter))

            elif line.startswith("TIMES") and "TIMES" in block_stack:
                block_stack.remove("TIMES")
                parameters.update(_read_times(fiter))

            elif line.startswith("HYSTE") and "HYSTE" in block_stack:
                block_stack.remove("HYSTE")
                parameters.update(_read_hyste(fiter))

            elif line.startswith("FOFT") and "FOFT" in block_stack:
                block_stack.remove("FOFT")
                oft, label_length = _read_oft(fiter, "FOFT", label_length)
                parameters.update(oft)

            elif line.startswith("COFT") and "COFT" in block_stack:
                block_stack.remove("COFT")
                oft, label_length = _read_oft(fiter, "COFT", label_length)
                parameters.update(oft)

            elif line.startswith("GOFT") and "GOFT" in block_stack:
                block_stack.remove("GOFT")
                oft, label_length = _read_oft(fiter, "GOFT", label_length)
                parameters.update(oft)

            elif line.startswith("ROFT") and "ROFT" in block_stack:
                block_stack.remove("ROFT")
                parameters.update(_read_roft(fiter))

            elif line.startswith("GENER") and "GENER" in block_stack:
                block_stack.remove("GENER")
                gener, flag, label_length = _read_gener(fiter, label_length, simulator)
                parameters.update(gener)

                if flag:
                    break

            elif line.startswith("TIMBC") and "TIMBC" in block_stack:
                block_stack.remove("TIMBC")
                parameters.update(_read_timbc(fiter))

            elif line.startswith("DIFFU") and "DIFFU" in block_stack:
                block_stack.remove("DIFFU")
                parameters.update(_read_diffu(fiter))

            elif line.startswith("OUTPT") and "OUTPT" in block_stack:
                block_stack.remove("OUTPT")
                outpt = _read_outpt(fiter)

                if "react" in parameters:
                    parameters["react"].update(outpt["react"])

                else:
                    parameters.update(outpt)

            elif line.startswith("OUTPU") and "OUTPU" in block_stack:
                block_stack.remove("OUTPU")
                parameters.update(_read_outpu(fiter))

            elif line.startswith("ELEME") and "ELEME" in block_stack:
                block_stack.remove("ELEME")
                eleme, label_length = _read_eleme(fiter, label_length)
                parameters.update(eleme)
                parameters["coordinates"] = False

            elif line.startswith("COORD") and "COORD" in block_stack:
                block_stack.remove("COORD")
                coord = _read_coord(fiter)

                for k, v in zip(parameters["elements"], coord):
                    parameters["elements"][k]["center"] = v

                parameters["coordinates"] = True

            elif line.startswith("CONNE") and "CONNE" in block_stack:
                block_stack.remove("CONNE")
                conne, flag, label_length = _read_conne(fiter, label_length)
                parameters.update(conne)

                if flag:
                    break

            elif line.startswith("INCON") and "INCON" in block_stack:
                block_stack.remove("INCON")
                incon, flag, label_length, n_variables = _read_incon(
                    fiter, label_length, n_variables, eos, simulator
                )
                parameters.update(incon)

                if flag:
                    break

            elif line.startswith("MESHM") and "MESHM" in block_stack:
                block_stack.remove("MESHM")
                parameters.update(_read_meshm(fiter))

            elif line.startswith("POISE") and "POISE" in block_stack:
                block_stack.remove("POISE")
                poise = _read_poise(fiter)

                if "react" in parameters:
                    parameters["react"].update(poise["react"])

                else:
                    parameters.update(poise)

            elif line.startswith("NOVER") and "NOVER" in block_stack:
                block_stack.remove("NOVER")
                parameters["nover"] = True

            elif line.startswith("ENDCY") and "ENDCY" in block_stack:
                block_stack.remove("ENDCY")
                end_comments = read_end_comments(fiter)

                if end_comments:
                    parameters["end_comments"] = end_comments

            # Stop reading if block stack is empty
            if not block_stack:
                break

    except:
        raise ReadError(f"failed to parse line {fiter.count}.")

    if flag:
        end_comments = read_end_comments(fiter)

        if end_comments:
            if isinstance(end_comments, str):
                end_comments = [end_comments]

            parameters["end_comments"] = ["+++", *end_comments]

    return parameters


def _read_dimen(f):
    """Read DIMEN block data."""
    fmt = block_to_format["DIMEN"]
    dimen = {"array_dimensions": {}}

    # Record 1
    line = f.next()
    data = read_record(line, fmt)

    dimen["array_dimensions"].update(
        {
            "n_rocks": data[0],
            "n_times": data[1],
            "n_generators": data[2],
            "n_rates": data[3],
            "n_increment_x": data[4],
            "n_increment_y": data[5],
            "n_increment_z": data[6],
            "n_increment_rad": data[7],
        }
    )

    # Record 2
    line = f.next()
    data = read_record(line, fmt)

    dimen["array_dimensions"].update(
        {
            "n_properties": data[0],
            "n_properties_times": data[1],
            "n_regions": data[2],
            "n_regions_parameters": data[3],
            "n_ltab": data[4],
            "n_rpcap": data[5],
            "n_elements_timbc": data[6],
            "n_timbc": data[7],
        }
    )

    return dimen


def _read_rocks(f, simulator="tough"):
    """Read ROCKS block data."""
    fmt = block_to_format["ROCKS"]
    rocks = {"rocks": {}}

    while True:
        line = f.next()

        if line.strip():
            # Record 1
            data = read_record(line, fmt[1])
            rock = data[0]
            rocks["rocks"][rock] = {
                "density": data[2],
                "porosity": data[3],
                "permeability": data[4] if len(set(data[4:7])) == 1 else data[4:7],
                "conductivity": data[7],
                "specific_heat": data[8],
            }

            nad = data[1] if data[1] else 0
            if nad:
                # Record 2
                line = f.next()
                data = read_record(line, fmt[2])
                rocks["rocks"][rock].update(
                    {
                        "compressibility": data[0],
                        "expansivity": data[1],
                        "conductivity_dry": data[2],
                        "tortuosity": data[3],
                        "klinkenberg_parameter": data[4],
                        "distribution_coefficient_3": data[5],
                        "distribution_coefficient_4": data[6],
                        "tortuosity_exponent": data[7],
                        "porosity_crit": data[8],
                    }
                )

            if nad >= 2:
                # TOUGHREACT
                if simulator == "toughreact" and nad >= 4:
                    line = f.next()
                    if line.strip():
                        rocks["rocks"][rock]["react_tp"] = read_model_record(
                            line, fmt[3]
                        )

                    line = f.next()
                    if line.strip():
                        rocks["rocks"][rock]["react_hcplaw"] = read_model_record(
                            line, fmt[4]
                        )

                rocks["rocks"][rock].update(_read_rpcap(f))

        else:
            break

    rocks["rocks"] = {k: prune_values(v) for k, v in rocks["rocks"].items()}

    return rocks


def _read_rpcap(f):
    """Read RPCAP block data."""
    fmt = block_to_format["RPCAP"]
    rpcap = {}

    for key in ["relative_permeability", "capillarity"]:
        line = f.next()
        if line.strip():
            rpcap[key] = read_model_record(line, fmt)

    return rpcap


def _read_react(f):
    """Read REACT block data."""
    fmt = block_to_format["REACT"]

    line = f.next()
    data = read_record(line, fmt)
    react = {
        "react": {
            "options": {i + 1: int(x) for i, x in enumerate(data[0]) if x.isdigit()}
        }
    }

    return react


def _read_flac(f, rocks_order):
    """Read FLAC block data."""
    fmt = block_to_format["FLAC"]
    flac = {"rocks": {}, "flac": {}}

    # Record 1
    line = f.next()
    data = read_record(line, fmt[1])
    flac["flac"]["creep"] = data[0]
    flac["flac"]["porosity_model"] = data[1]
    flac["flac"]["version"] = data[2]

    # Additional records
    for rock in rocks_order:
        flac["rocks"][rock] = {}

        line = f.next()
        data = read_record(line, fmt[2])
        flac["rocks"][rock]["permeability_model"] = {
            "id": data[0],
            "parameters": prune_values(data[1:]),
        }

        line = f.next()
        data = read_record(line, fmt[3])
        flac["rocks"][rock]["equivalent_pore_pressure"] = {
            "id": data[0],
            "parameters": prune_values(data[2:]),
        }

    flac["flac"] = prune_values(flac["flac"])

    return flac


def _read_chemp(f):
    """Read CHEMP block data."""
    fmt = block_to_format["CHEMP"]
    chemp = {"chemical_properties": {}}

    # Record 1
    line = f.next()
    n = read_record(line, fmt[1])[0]

    # Record 2
    for _ in range(n):
        tmp = {}

        line = f.next()
        data = read_record(line, fmt[2])
        chem = data[0]

        line = f.next()
        data = read_record(line, fmt[3])
        tmp["temperature_crit"] = data[0]
        tmp["pressure_crit"] = data[1]
        tmp["compressibility_crit"] = data[2]
        tmp["pitzer_factor"] = data[3]
        tmp["dipole_moment"] = data[4]

        line = f.next()
        data = read_record(line, fmt[3])
        tmp["boiling_point"] = data[0]
        tmp["vapor_pressure_a"] = data[1]
        tmp["vapor_pressure_b"] = data[2]
        tmp["vapor_pressure_c"] = data[3]
        tmp["vapor_pressure_d"] = data[4]

        line = f.next()
        data = read_record(line, fmt[3])
        tmp["molecular_weight"] = data[0]
        tmp["heat_capacity_a"] = data[1]
        tmp["heat_capacity_b"] = data[2]
        tmp["heat_capacity_c"] = data[3]
        tmp["heat_capacity_d"] = data[4]

        line = f.next()
        data = read_record(line, fmt[3])
        tmp["napl_density_ref"] = data[0]
        tmp["napl_temperature_ref"] = data[1]
        tmp["gas_diffusivity_ref"] = data[2]
        tmp["gas_temperature_ref"] = data[3]
        tmp["exponent"] = data[4]

        line = f.next()
        data = read_record(line, fmt[3])
        tmp["napl_viscosity_a"] = data[0]
        tmp["napl_viscosity_b"] = data[1]
        tmp["napl_viscosity_c"] = data[2]
        tmp["napl_viscosity_d"] = data[3]
        tmp["volume_crit"] = data[4]

        line = f.next()
        data = read_record(line, fmt[3])
        tmp["solubility_a"] = data[0]
        tmp["solubility_b"] = data[1]
        tmp["solubility_c"] = data[2]
        tmp["solubility_d"] = data[3]

        line = f.next()
        data = read_record(line, fmt[3])
        tmp["oc_coeff"] = data[0]
        tmp["oc_fraction"] = data[1]
        tmp["oc_decay"] = data[2]

        chemp["chemical_properties"][chem] = tmp

    return chemp


def _read_ncgas(f):
    """Read NCGAS block data."""
    fmt = block_to_format["NCGAS"]
    ncgas = {"non_condensible_gas": []}

    # Record 1
    line = f.next()
    n = read_record(line, fmt[1])[0]

    # Record 2
    for _ in range(n):
        line = f.next()
        data = read_record(line, fmt[2])
        ncgas["non_condensible_gas"].append(data[0])

    return ncgas


def _read_multi(f):
    """Read MULTI block data."""
    fmt = block_to_format["MULTI"]
    multi = {}

    line = f.next()
    data = read_record(line, fmt)
    multi["n_component"] = data[0]
    multi["isothermal"] = data[1] == data[0]
    multi["n_phase"] = data[2]
    multi["do_diffusion"] = data[3] == 8
    multi["n_component_incon"] = data[4]

    return multi


def _read_solvr(f):
    """Read SOLVR block data."""
    fmt = block_to_format["SOLVR"]
    solvr = {}

    line = f.next()
    data = read_record(line, fmt)
    solvr["solver"] = {
        "method": data[0],
        "z_precond": data[2],
        "o_precond": data[4],
        "rel_iter_max": data[5],
        "eps": data[6],
    }

    return solvr


def _read_param(f, n_variables, eos=None):
    """Read PARAM block data."""
    fmt = block_to_format["PARAM"]
    param = {}

    # Record 1
    line = f.next()
    data = read_record(line, fmt[1])
    param["options"] = {
        "n_iteration": data[0],
        "verbosity": data[1],
        "n_cycle": data[2],
        "n_second": data[3],
        "n_cycle_print": data[4],
        "temperature_dependence_gas": data[7],
        "effective_strength_vapor": data[8],
    }
    param["extra_options"] = {
        i + 1: int(x) for i, x in enumerate(data[5]) if x.isdigit()
    }

    # Record 2
    line = f.next()
    data = read_record(line, fmt[2])
    param["options"].update(
        {
            "t_ini": data[0],
            "t_max": data[1],
            "t_steps": data[2],
            "t_step_max": data[3],
            "gravity": data[6],
            "t_reduce_factor": data[7],
            "mesh_scale_factor": data[8],
        }
    )
    wdata = data[4]

    t_steps = data[2]
    if t_steps:
        if t_steps >= 0.0:
            param["options"]["t_steps"] = t_steps

        else:
            t_steps = int(-t_steps)
            param["options"]["t_steps"] = []

            for _ in range(t_steps):
                line = f.next()
                data = read_record(line, fmt[3])
                param["options"]["t_steps"] += prune_values(data)

            if len(param["options"]["t_steps"]) == 1:
                param["options"]["t_steps"] = param["options"]["t_steps"][0]

    # TOUGHREACT
    if wdata == "wdata":
        line = f.next()
        n = int(line.strip().split()[0])

        if n:
            param["options"]["react_wdata"] = []
            for _ in range(n):
                line = f.next()
                param["options"]["react_wdata"].append(line.strip()[:5])

    # Record 3
    line = f.next()
    data = read_record(line, fmt[4])
    param["options"].update(
        {
            "eps1": data[0],
            "eps2": data[1],
            "w_upstream": data[3],
            "w_newton": data[4],
            "derivative_factor": data[5],
        }
    )

    # Record 4 (ECO2M, TMVOC)
    if eos in {"eco2m", "tmvoc"}:
        line = f.next()
        data = read_record(line, "5d")
        param["default"] = {"phase_composition": data[0]}

    # Record 5
    data = read_primary_variables(f, fmt[5], n_variables)

    if "default" not in param:
        param["default"] = {}

    if any(x is not None for x in data):
        data = prune_values(data)
        param["default"]["initial_condition"] = data

    if not n_variables:
        n_variables = len(data)

    # Remove Nones
    param["options"] = prune_values(param["options"])
    param["extra_options"] = prune_values(param["extra_options"])

    return param, n_variables


def _read_selec(f):
    """Read SELEC block data."""
    fmt = block_to_format["SELEC"]
    selec = {"selections": {}}

    line = f.next()
    data = read_record(line, fmt[1])
    selec["selections"]["integers"] = {k + 1: v for k, v in enumerate(data)}

    if selec["selections"]["integers"][1]:
        selec["selections"]["floats"] = []
        for _ in range(selec["selections"]["integers"][1]):
            line = f.next()
            data = read_record(line, fmt[2])
            selec["selections"]["floats"].append(prune_values(data))

    selec["selections"]["integers"] = prune_values(selec["selections"]["integers"])
    if selec["selections"]["integers"][1] == 1:
        selec["selections"]["floats"] = selec["selections"]["floats"][0]

    return selec


def _read_indom(f, n_variables, eos=None):
    """Read INDOM block data."""
    fmt = block_to_format["INDOM"]
    indom = {"rocks": {}}

    line = f.next()
    while True:
        if line.strip():
            # Record 1
            data = read_record(line, fmt[5])
            rock = data[0]
            phase_composition = data[1]  # ECO2M, TMVOC

            # Record 2
            data = read_primary_variables(f, fmt[0], n_variables)
            data = prune_values(data)
            indom["rocks"][rock] = {"initial_condition": data}

            if not n_variables:
                n_variables = len(data)

            if eos in {"eco2m", "tmvoc"}:
                indom["rocks"][rock]["phase_composition"] = phase_composition

        else:
            break

        line = f.next()

    return indom, n_variables


def _read_momop(f):
    """Read MOMOP block data."""
    fmt = block_to_format["MOMOP"]

    line = f.next()
    data = read_record(line, fmt)
    momop = {
        "more_options": {i + 1: int(x) for i, x in enumerate(data[0]) if x.isdigit()}
    }

    return momop


def _read_times(f):
    """Read TIMES block data."""
    fmt = block_to_format["TIMES"]
    times = {"times": []}

    # Record 1
    line = f.next()
    data = read_record(line, fmt[1])
    n_times = data[0]

    # Record 2
    while len(times["times"]) < n_times:
        line = f.next()
        data = read_record(line, fmt[2])
        times["times"] += prune_values(data)

    return times


def _read_hyste(f):
    """Read HYSTE block data."""
    fmt = block_to_format["HYSTE"]
    hyste = {"hysteresis_options": {}}

    line = f.next()
    data = read_record(line, fmt)
    hyste["hysteresis_options"] = {k + 1: v for k, v in enumerate(data)}
    hyste["hysteresis_options"] = prune_values(hyste["hysteresis_options"])

    return hyste


def _read_oft(f, oft, label_length):
    """Read FOFT, COFT and GOFT blocks data."""
    key = oft_to_key[oft]
    fmt = block_to_format[oft]
    history = {key: []}

    line = f.next()
    if not label_length:
        label_length = get_label_length(line[:9])
    label_format = f"{{:>{label_length if oft != 'COFT' else 2 * label_length}}}"

    while True:
        if line.strip():
            data = read_record(line, fmt[label_length])
            history[key].append(label_format.format(data[0]))

        else:
            break

        line = f.next()

    return history, label_length


def _read_roft(f):
    """Read ROFT block data."""
    fmt = block_to_format["ROFT"]
    history = {"rock_history": []}

    line = f.next()
    while True:
        if line.strip():
            data = read_record(line, fmt)
            rock1 = data[0] if data[0] else ""
            rock2 = data[1] if data[1] else ""
            history["rock_history"].append([rock1, rock2])

        else:
            break

        line = f.next()

    return history


def _read_gener(f, label_length, simulator="tough"):
    """Read GENER block data."""

    def read_table(f, n, fmt):
        table = []
        while len(table) < n:
            line = f.next()
            data = read_record(line, fmt)
            table += prune_values(data)

        return table

    fmt = block_to_format["GENER"]
    gener = {"generators": []}

    line = f.next()
    if not label_length:
        label_length = get_label_length(line[:9])
    label_format = f"{{:>{label_length}}}"

    flag = False
    while True:
        if line.strip() and not line.startswith("+++"):
            data = read_record(line, fmt[label_length])
            tmp = {
                "label": label_format.format(data[0]),
                "name": data[1],
                "nseq": data[2],
                "nadd": data[3],
                "nads": data[4],
                "type": data[7],
                "layer_thickness": data[11],
            }
            ktab = data[12]  # TOUGHREACT

            ltab = data[5]
            if ltab and ltab > 1 and tmp["type"] != "DELV":
                itab = data[8]
                keys = ["times", "rates"]
                keys += (
                    ["specific_enthalpy"] if itab else []
                )  # Specific enthalpy must be provided for time dependent injection
                for key in keys:
                    tmp[key] = read_table(f, ltab, fmt[0])

            else:
                tmp.update(
                    {
                        "times": None,
                        "rates": data[9],
                        "specific_enthalpy": data[10],
                    }
                )

            if ltab and tmp["type"] == "DELV":
                tmp["n_layer"] = ltab

            if simulator == "toughreact" and ktab:
                tmp["conductivity_times"] = read_table(f, ktab, fmt[0])
                tmp["conductivity_factors"] = read_table(f, ktab, fmt[0])

            gener["generators"].append(tmp)

        else:
            flag = line.startswith("+++")
            break

        line = f.next()

    return (
        {"generators": [prune_values(generator) for generator in gener["generators"]]},
        flag,
        label_length,
    )


def _read_timbc(f):
    """Read TIMBC block data."""
    timbc = {"boundary_conditions": []}

    # Record 1
    line = f.next().strip()
    ntptab = int(line)

    for _ in range(ntptab):
        # Record 2
        line = f.next().strip()
        data = [int(x) for x in line.split()]
        if len(data) < 2:
            raise ReadError()

        nbcp = data[0]
        nbcpv = data[1]

        # Record 3
        bcelm = f.next().strip()

        # Record 4
        times = []
        values = []
        for _ in range(nbcp):
            line = f.next().strip()
            data = [float(x) for x in line.split()]
            times.append(data[0])
            values.append(data[1])

        tmp = {
            "label": bcelm,
            "variable": nbcpv,
            "times": times,
            "values": values,
        }
        timbc["boundary_conditions"].append(tmp)

    return timbc


def _read_diffu(f):
    """Read DIFFU block data."""
    fmt = block_to_format["DIFFU"]
    diffu = {"diffusion": []}

    while True:
        i = f.tell()
        line = f.next()

        if line.split():
            try:
                data = read_record(line, fmt)
                diffu["diffusion"].append(prune_values(data))
            except ValueError:
                f.seek(i, increment=-1)
                break
        else:
            break

    return diffu


def _read_outpt(f):
    """Read OUTPT block data."""
    outpt = {"react": {"output": {}}}

    line = f.next().strip()
    data = [int(x) for x in line.split()]  # Free-format

    if len(data):
        outpt["react"]["output"]["format"] = data[0]

    if len(data) > 1:
        outpt["react"]["output"]["shape"] = data[1:]

    return outpt


def _read_outpu(f):
    """Read OUTPU block data."""
    fmt = block_to_format["OUTPU"]
    outpu = {"output": {}}

    # Format
    line = f.next().strip()
    if line and not line.isdigit():
        outpu["output"]["format"] = line if line else None
        line = f.next().strip()

    # Variables
    if line.isdigit():
        num_vars = int(line)
        outpu["output"]["variables"] = []

        for _ in range(num_vars):
            line = f.next()
            data = read_record(line, fmt[3])
            name = data[0].lower()

            tmp = prune_values(data[1:])
            options = None if len(tmp) == 0 else tmp[0] if len(tmp) == 1 else tmp

            outpu["output"]["variables"].append({"name": name})
            if options is not None:
                outpu["output"]["variables"][-1]["options"] = options

    return outpu


def _read_eleme(f, label_length):
    """Read ELEME block data."""
    fmt = block_to_format["ELEME"]
    eleme = {"elements": {}}

    line = f.next()
    if not label_length:
        label_length = get_label_length(line[:9])
    label_format = f"{{:>{label_length}}}"

    while True:
        if line.strip():
            data = read_record(line, fmt[label_length])
            label = label_format.format(data[0])
            label = label.lstrip() if label.lstrip().isalpha() else label
            rock = data[3]
            if rock:
                rock = rock.strip()
                rock = int(rock) if rock.isdigit() else rock
            eleme["elements"][label] = {
                "nseq": data[1],
                "nadd": data[2],
                "material": rock,
                "volume": data[4],
                "heat_exchange_area": data[5],
                "permeability_modifier": data[6],
                "center": data[7:10],
            }

        else:
            break

        line = f.next()

    eleme["elements"] = {k: prune_values(v) for k, v in eleme["elements"].items()}

    return eleme, label_length


def _read_coord(f):
    """Read COORD block data."""
    fmt = block_to_format["COORD"]
    coord = []

    line = f.next()
    while True:
        if line.strip():
            data = read_record(line, fmt)
            coord.append(data)
        else:
            break

        line = f.next()

    return coord


def _read_conne(f, label_length):
    """Read CONNE block data."""
    fmt = block_to_format["CONNE"]
    conne = {"connections": {}}

    line = f.next()
    if not label_length:
        label_length = get_label_length(line[:9])
    label_format = f"{{:>{2 * label_length}}}"

    flag = False
    while True:
        if line.strip() and not line.startswith("+++"):
            data = read_record(line, fmt[label_length])
            label = label_format.format(data[0])
            conne["connections"][label] = {
                "nseq": data[1],
                "nadd": data[2:4],
                "permeability_direction": data[4],
                "nodal_distances": data[5:7],
                "interface_area": data[7],
                "gravity_cosine_angle": data[8],
                "radiant_emittance_factor": data[9],
            }

        else:
            flag = line.startswith("+++")
            break

        line = f.next()

    conne["connections"] = {k: prune_values(v) for k, v in conne["connections"].items()}

    return conne, flag, label_length


def _read_incon(f, label_length, n_variables, eos=None, simulator="tough"):
    """Read INCON block data."""
    fmt = block_to_format["INCON"]
    fmt2 = (
        fmt[simulator]
        if simulator == "toughreact"
        else fmt[eos]
        if eos in fmt
        else fmt["default"]
    )
    incon = {"initial_conditions": {}}

    line = f.next()
    if not label_length:
        label_length = get_label_length(line[:9])
    label_format = f"{{:>{label_length}}}"

    flag = False
    while True:
        if line.strip() and not line.startswith("+++"):
            # Record 1
            data = read_record(line, fmt2[label_length])
            label = label_format.format(data[0])
            incon["initial_conditions"][label] = {"porosity": data[3]}

            if simulator == "toughreact":
                permeability = data[4] if len(set(data[4:7])) == 1 else data[4:7]
                incon["initial_conditions"][label]["permeability"] = (
                    permeability if permeability else None
                )

            elif eos in {"eco2m", "tmvoc"}:
                incon["initial_conditions"][label]["phase_composition"] = data[4]

            else:
                userx = prune_values(data[4:9])
                incon["initial_conditions"][label]["userx"] = userx if userx else None

            # Record 2
            data = read_primary_variables(f, fmt[0], n_variables)
            data = prune_values(data)
            incon["initial_conditions"][label]["values"] = data

            if not n_variables:
                n_variables = len(data)

        else:
            flag = line.startswith("+++")
            break

        line = f.next()

    incon["initial_conditions"] = {
        k: prune_values(v) for k, v in incon["initial_conditions"].items()
    }

    return incon, flag, label_length, n_variables


def _read_meshm(f):
    """Read MESHM block data."""

    def read_minc(f, data, fmt):
        """Read MINC data."""
        minc = {}

        # Record 1
        line = f.next()
        data = read_record(line, fmt[1])
        minc["type"] = data[1].lower()
        minc["dual"] = data[3].lower()

        # Record 2
        line = f.next()
        data = read_record(line, fmt[2])
        minc["n_minc"] = data[0]
        n_volume = data[1]
        minc["where"] = data[2].lower()
        minc["parameters"] = prune_values(data[3:])

        # Record 3
        minc["volumes"] = []

        while len(minc["volumes"]) < n_volume:
            line = f.next()
            data = read_record(line, fmt[3])
            minc["volumes"] += prune_values(data)

        return minc

    fmt = block_to_format["MESHM"]
    fmt_minc = fmt["MINC"]

    # Mesh type
    line = f.next().strip()
    data = read_record(line, fmt[1])
    mesh_type = data[0].upper()

    if mesh_type in {"XYZ", "RZ2D", "RZ2DL"}:
        meshm = {"meshmaker": {"type": mesh_type.lower()}}

    else:
        meshm = {"minc": {}}

    # XYZ
    if mesh_type == "XYZ":
        fmt = fmt["XYZ"]

        # Record 1
        line = f.next().strip()
        data = read_record(line, fmt[1])
        meshm["meshmaker"]["angle"] = data[0]

        # Record 2
        meshm["meshmaker"]["parameters"] = []

        while True:
            line = f.next()

            if line.strip():
                data = read_record(line, fmt[2])
                tmp = {
                    "type": data[0].lower(),
                    "n_increment": data[1],
                }

                if data[2]:
                    tmp["sizes"] = data[2]

                else:
                    sizes = []
                    while len(sizes) < tmp["n_increment"]:
                        line = f.next()
                        data = read_record(line, fmt[3])
                        sizes += prune_values(data)

                    tmp["sizes"] = sizes[: tmp["n_increment"]]

                meshm["meshmaker"]["parameters"].append(tmp)

            else:
                break

    # RZ2D
    elif mesh_type in {"RZ2D", "RZ2DL"}:
        fmt = fmt["RZ2D"]

        # Record 1
        meshm["meshmaker"]["parameters"] = []

        while True:
            line = f.next()

            if line.strip():
                data = read_record(line, fmt[1])
                data_type = data[0].upper()

                if data_type == "RADII":
                    line = f.next()
                    data = read_record(line, fmt["RADII"][1])
                    n = data[0]

                    radii = []
                    while len(radii) < n:
                        line = f.next()
                        data = read_record(line, fmt["RADII"][2])
                        radii += prune_values(data)

                    tmp = {
                        "type": data_type.lower(),
                        "radii": radii[:n],
                    }
                    meshm["meshmaker"]["parameters"].append(tmp)

                elif data_type == "EQUID":
                    line = f.next()
                    data = read_record(line, fmt["EQUID"])

                    tmp = {
                        "type": data_type.lower(),
                        "n_increment": data[0],
                        "size": data[2],
                    }
                    meshm["meshmaker"]["parameters"].append(tmp)

                elif data_type == "LOGAR":
                    line = f.next()
                    data = read_record(line, fmt["LOGAR"])

                    tmp = {
                        "type": data_type.lower(),
                        "n_increment": data[0],
                        "radius": data[2],
                        "radius_ref": data[3],
                    }
                    meshm["meshmaker"]["parameters"].append(tmp)

                elif data_type == "LAYER":
                    # Record 1
                    line = f.next()
                    data = read_record(line, fmt["LAYER"][1])
                    n = data[0]

                    # Record 2
                    thicknesses = []
                    while len(thicknesses) < n:
                        line = f.next()
                        data = read_record(line, fmt["LAYER"][2])
                        thicknesses += prune_values(data)

                    tmp = {"type": data_type.lower(), "thicknesses": thicknesses[:n]}
                    meshm["meshmaker"]["parameters"].append(tmp)

                    # LAYER closes the block RZ2D
                    # RZ2D can be followed by MINC
                    line = f.next()

                    if line.startswith("MINC"):
                        meshm["minc"] = read_minc(f, data, fmt_minc)

                    else:
                        break

            else:
                break

    # MINC
    elif mesh_type == "MINC":
        meshm["minc"] = read_minc(f, data, fmt_minc)

    return meshm


def _read_poise(f):
    """Read POISE block data."""
    poise = {"react": {"poiseuille": {}}}

    line = f.next().strip()
    data = [float(x) for x in line.split()]  # Free-format

    if len(data) < 5:
        raise ReadError()

    poise["react"]["poiseuille"]["start"] = data[:2]
    poise["react"]["poiseuille"]["end"] = data[2:4]
    poise["react"]["poiseuille"]["aperture"] = data[4]

    return poise



# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import os
import sys

sys.path.insert(0, os.path.abspath("."))
sys.path.insert(0, os.path.abspath("../"))
sys.path.insert(0, os.path.abspath(".../"))


# -- Project information -----------------------------------------------------

import toughio

project = "toughio"
copyright = "2023, Keurfon Luu"
author = "Keurfon Luu"

version = toughio.__version__
release = version


# -- General configuration ---------------------------------------------------

master_doc = "index"

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
    "sphinx.ext.mathjax",
    "sphinx.ext.ifconfig",
    "sphinx.ext.viewcode",
    "sphinx.ext.githubpages",
    "sphinx.ext.napoleon",
    "sphinxarg.ext",
    "sphinxcontrib.bibtex",
    "sphinx_gallery.gen_gallery",
    "sphinx_subfigure",
]

# Sphinx Gallery settings
from sphinx_gallery.sorting import FileNameSortKey

sphinx_gallery_conf = {
    "examples_dirs": ["../../examples/"],
    "gallery_dirs": ["examples/"],
    "filename_pattern": r"\.py",
    "download_all_examples": False,
    "within_subsection_order": FileNameSortKey,
    "backreferences_dir": None,
    "doc_module": "toughio",
    "image_scrapers": (
        "matplotlib",
        "pyvista",
    ),
    "first_notebook_cell": (
        "%matplotlib inline\n"
    ),
}

# PyVista settings
import pyvista

pyvista.BUILDING_GALLERY = True

# Napoleon settings
napoleon_google_docstring = False
napoleon_numpy_docstring = True
napoleon_include_init_with_doc = False
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = False
# napoleon_use_admonition_for_examples = False
# napoleon_use_admonition_for_notes = False
# napoleon_use_admonition_for_references = False
# napoleon_use_ivar = True
# napoleon_use_param = True
# napoleon_use_rtype = True

# Numfig settings
numfig = True
numfig_format = {
    "figure": "Figure %s",
}

# BibTeX
bibtex_bibfiles = ["references.bib"]

# Add any paths that contain templates here, relative to this directory.
templates_path = [
    "_templates",
]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = [
    "_build",
    "Thumbs.db",
    ".DS_Store",
]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True

# Class documentation
autoclass_content = "both"


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "pydata_sphinx_theme"
html_theme_options = {
   "logo": {
      "image_light": "_static/logo-light.png",
      "image_dark": "_static/logo.png",
   }
}
html_theme_path = [
    "_themes",
]
html_context = {
    "default_mode": "light",
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

html_css_files = [
    "css/custom.css",
]


import logging

from ..._common import prune_values
from .._common import to_str


def getval(parameters, keys, default):
    """Get key in parameters dict or return default value."""
    try:
        if isinstance(keys, str):
            return parameters[keys]

        else:
            value = parameters
            for key in keys:
                value = value[key]

            return value

    except KeyError:
        key_str = (
            f"'{keys}'"
            if isinstance(keys, str)
            else "".join(f"['{key}']" for key in keys)
        )
        logging.warning(f"Key {key_str} is not defined. Setting to {default}.")

        return default


def write_ffrecord(
    values,
    verbose=False,
    fmt=None,
    int_fmt="{:4d}",
    float_fmt="{{:9f}}",
    str_fmt="{:20}",
    end="",
):
    """Write free-format record."""
    if verbose:
        if fmt:
            values = [to_str(value, f) for value, f, in zip(values, fmt)]
            return [f"{' '.join(values)}{end}"]

        else:
            return [
                f"{' '.join(to_str(value, int_fmt if isinstance(value, int) else float_fmt if isinstance(value, float) else str_fmt) for value in values)}{end}"
            ]

    else:
        return [f"{' '.join(str(x) for x in values)}{end}"]


def read_end_comments(fiter):
    # Save end comments
    end_comments = []
    while True:
        try:
            end_comments.append(fiter.next().rstrip())

        except StopIteration:
            break

    # Remove trailing empty records
    end_comments = [comment if comment else None for comment in end_comments]
    end_comments = prune_values(end_comments)
    if end_comments:
        return (
            end_comments[0]
            if len(end_comments) == 1
            else [comment if comment else "" for comment in end_comments]
        )

    else:
        return None



from .h5 import write as write_h5
from .input import read as read_input
from .input import register as register_input
from .input import write as write_input
from .output import Output
from .output import read as read_output
from .output import register as register_output
from .output import write as write_output
from .table import read as read_table
from .table import register as register_table

__all__ = [
    "Output",
    "register_input",
    "register_output",
    "read_input",
    "write_h5",
    "write_input",
    "read_output",
    "write_output",
    "read_table",
    "register_table",
]



import numpy as np

from ...input import tough
from .._common import Output, reorder_labels

__all__ = [
    "read",
]


def read(filename, file_type, labels_order=None):
    """
    Read SAVE file.

    Parameters
    ----------
    filename : str, pathlike or buffer
        Input file name or buffer.
    file_type : str
        Input file type.
    labels_order : list of array_like
        List of labels. If None, output will be assumed ordered.

    Returns
    -------
    namedtuple or list of namedtuple
        namedtuple (type, format, time, labels, data) or list of namedtuple for each time step.

    """
    parameters = tough.read(filename)

    labels = list(parameters["initial_conditions"])
    variables = [v["values"] for v in parameters["initial_conditions"].values()]

    data = {f"X{i + 1}": x for i, x in enumerate(np.transpose(variables))}

    data["porosity"] = np.array(
        [v["porosity"] for v in parameters["initial_conditions"].values()]
    )

    userx = [
        v["userx"] for v in parameters["initial_conditions"].values() if "userx" in v
    ]
    if userx:
        data["userx"] = np.array(userx)

    labels_order = labels_order if labels_order else parameters["initial_conditions"]
    output = Output(file_type, None, np.array(labels), data)
    return reorder_labels(output, labels_order)



"""
Import and visualize simulation outputs in PyVista
==================================================

Once the TOUGH simulation has ended, there is not much things left to do with :mod:`toughio`. We can either:

-   Run the command :code:`toughio-export OUTPUT_ELEME.csv -m mesh.pickle` and drag-and-drop the generated VTK file in a compatible visualization software (e.g., ParaView),

.. image:: ../../figures/co2_leakage_along_a_fault_paraview.png

-   Import the output file *OUTPUT_ELEME.csv* in Python using :mod:`toughio` (two-liners) and do whatever post-processing we want and need.

This example briefly shows how to import TOUGH output file using :mod:`toughio` and how to visualize it with :mod:`pyvista` (adapted from :mod:`pyvista`'s `documentation <https://docs.pyvista.org/>`__).

"""

########################################################################################
# First, we unpickle the mesh file and import the last time step of the output file (`OUTPUT_ELEME.csv` in TOUGH3, `OUTPUT` in TOUGH2).

import toughio

mesh = toughio.read_mesh("mesh.pickle")
mesh.read_output("OUTPUT_ELEME.csv", time_step=-1)

########################################################################################

########################################################################################
# Once output data have been imported, we can simply call the method :meth:`toughio.Mesh.to_pyvista` to convert the mesh to a :class:`pyvista.UnstructuredGrid` object that can be processed by :mod:`pyvista`.

import pyvista
pyvista.set_plot_theme("document")

p = pyvista.Plotter(window_size=(1000, 1000))
p.add_mesh(
    mesh.to_pyvista(),
    scalars="SAT_G",
    cmap="viridis_r",
    clim=(0.0, 1.0),
    n_colors=20,
    show_edges=False,
    edge_color=(0.5, 0.5, 0.5),
    scalar_bar_args={
        "title": "Gas saturation",
        "position_y": 0.01,
        "vertical": False,
        "n_labels": 6,
        "fmt": "%.1f",
        "title_font_size": 20,
        "font_family": "arial",
        "shadow": True,
    },
)
p.show_grid(
    show_xaxis=True,
    show_yaxis=False,
    show_zaxis=True,
    xlabel="Distance (m)",
    zlabel="Elevation (m)",
    ticks="outside",
    font_family="arial",
)
p.view_xz()
p.show()

########################################################################################



import os
import tempfile
from collections.abc import Sequence

import numpy as np

import toughio

np.random.seed(42)

tet_mesh = toughio.Mesh(
    points=np.array(
        [
            [0.0, 0.0, 0.0],
            [1.0, 0.0, 0.0],
            [1.0, 1.0, 0.0],
            [0.0, 1.0, 0.0],
            [0.5, 0.5, 0.5],
        ]
    ),
    cells=[("tetra", np.array([[0, 1, 2, 4], [0, 2, 3, 4]]))],
    point_data={"a": np.random.rand(5), "b": np.random.rand(5)},
    cell_data={"c": np.random.rand(2), "material": np.ones(2, dtype=np.int64)},
)

hex_mesh = toughio.Mesh(
    points=np.array(
        [
            [0.0, 0.0, 0.0],
            [1.0, 0.0, 0.0],
            [1.0, 1.0, 0.0],
            [0.0, 1.0, 0.0],
            [0.0, 0.0, 1.0],
            [1.0, 0.0, 1.0],
            [1.0, 1.0, 1.0],
            [0.0, 1.0, 1.0],
        ]
    ),
    cells=[("hexahedron", np.array([[0, 1, 2, 3, 4, 5, 6, 7]]))],
    point_data={"a": np.random.rand(8), "b": np.random.rand(8)},
    cell_data={"c": np.random.rand(1), "material": np.ones(1, dtype=np.int64)},
)

hybrid_mesh = toughio.Mesh(
    points=np.array(
        [
            [0.0, 0.0, 0.0],
            [1.0, 0.0, 0.0],
            [1.0, 1.0, 0.0],
            [0.0, 1.0, 0.0],
            [0.0, 0.0, 1.0],
            [1.0, 0.0, 1.0],
            [1.0, 1.0, 1.0],
            [0.0, 1.0, 1.0],
            [0.5, 0.5, 1.5],
            [0.0, 0.5, 1.5],
            [1.0, 0.5, 1.5],
            [2.0, 0.0, 0.0],
            [2.0, 1.0, 0.0],
            [-1.0, 0.0, 0.0],
            [-1.0, 1.0, 0.0],
        ]
    ),
    cells=[
        ("hexahedron", np.array([[0, 1, 2, 3, 4, 5, 6, 7]])),
        ("pyramid", np.array([[4, 5, 6, 7, 8]])),
        ("tetra", np.array([[4, 8, 7, 9], [5, 6, 8, 10]])),
        ("wedge", np.array([[1, 11, 5, 2, 12, 6], [13, 0, 4, 14, 3, 7]])),
    ],
    point_data={"a": np.random.rand(15), "b": np.random.rand(15)},
    cell_data={"c": np.random.rand(6), "material": np.ones(6, dtype=np.int64)},
)

output_eleme = [
    toughio.Output(
        "element",
        float(time),
        np.array([f"AAA0{i}" for i in range(10)]),
        {
            "X": np.random.rand(10),
            "Y": np.random.rand(10),
            "Z": np.random.rand(10),
            "PRES": np.random.rand(10),
            "TEMP": np.random.rand(10),
        },
    )
    for time in range(3)
]

output_conne = [
    toughio.Output(
        "connection",
        float(time),
        np.array([[f"AAA0{i}", f"AAA0{i}"] for i in range(10)]),
        {
            "X": np.random.rand(10),
            "Y": np.random.rand(10),
            "Z": np.random.rand(10),
            "HEAT": np.random.rand(10),
            "FLOW": np.random.rand(10),
        },
    )
    for time in range(3)
]


def tempdir(filename=None):
    temp_dir = tempfile.mkdtemp()
    return os.path.join(temp_dir, filename) if filename else temp_dir


def write_read(filename, obj, writer, reader, writer_kws=None, reader_kws=None):
    writer_kws = writer_kws if writer_kws else {}
    reader_kws = reader_kws if reader_kws else {}

    filepath = tempdir(filename)
    if obj is not None:
        writer(filepath, obj, **writer_kws)
    else:
        writer(filepath, **writer_kws)

    return reader(filepath, **reader_kws)


def random_string(n):
    import random
    from string import ascii_lowercase

    return "".join(random.choice(ascii_lowercase) for _ in range(n))


def random_label(label_length=5):
    n = label_length - 3
    fmt = f"{{:0{n}d}}"

    return random_string(3) + fmt.format(np.random.randint(10**n))


def allclose(x, y, atol=1.0e-8, ignore_keys=None, ignore_none=False):
    ignore_keys = ignore_keys if ignore_keys is not None else []

    if isinstance(x, dict):
        assert isinstance(y, dict)

        for k, v in x.items():
            if k in ignore_keys:
                continue

            if ignore_none and v is None:
                continue

            try:
                assert allclose(v, y[k], atol=atol, ignore_none=ignore_none)

            except KeyError as e:
                print("x =", v, "\ny =", y[k], "\n")
                raise KeyError(e)

    else:
        try:
            if isinstance(x, toughio.Mesh):
                assert isinstance(y, toughio.Mesh)

                assert allclose(x.points, y.points, atol=atol)
                assert allclose(x.cells, y.cells, atol=atol)

                if x.point_data:
                    assert allclose(x.point_data, y.point_data, atol=atol)

                if x.cell_data:
                    assert allclose(x.cell_data, y.cell_data, atol=atol)

            elif isinstance(x, toughio.Output):
                assert isinstance(y, toughio.Output)

                assert allclose(x.type, y.type, atol=atol)
                assert allclose(x.time, y.time, atol=atol)
                assert allclose(x.data, y.data, atol=atol)

                if np.ndim(x.labels) != 0:
                    assert allclose(x.labels, y.labels, atol=atol)

            # str is a Sequence
            elif isinstance(x, (str, type(None))):
                assert x == y

            elif isinstance(x, (Sequence, np.ndarray)):
                for xx, yy in zip(x, y):
                    assert allclose(xx, yy, atol=atol, ignore_none=ignore_none)

            else:
                assert np.allclose(x, y, atol=atol)

        except Exception as e:
            print("x =", x, "\ny =", y, "\n")
            raise Exception(e)

    return True


def convert_outputs_labels(outputs, connection=False):
    for output in outputs:
        try:
            if not connection:
                output.labels[:] = toughio.convert_labels(output.labels)

            else:
                labels = toughio.convert_labels(output.labels.ravel())
                output.labels[:] = labels.reshape((labels.size // 2, 2))

        except TypeError:
            pass
